{
  "columns": [
    {
      "aid": "54",
      "attribute": "Name"
    },
    {
      "aid": "54",
      "attribute": "Factor"
    },
    {
      "aid": "54",
      "attribute": "Offset"
    },
    {
      "aid": "47",
      "attribute": "Name"
    },
    {
      "aid": "47",
      "attribute": "Length"
    },
    {
      "aid": "47",
      "attribute": "Mass"
    }
  ],
  "joins": [
    {
      "aidFrom": "54",
      "aidTo": "47",
      "relation": "PhysDimension"
    }
  ]
}